
#ifndef ESC_HPP
#define ESC_HPP

#include "misc.hpp"

#include <vector>
#include <utility>
#include <set>
#include <map>
#include <cmath>
#include <iomanip>
#include <algorithm>
/* The followings are included in esc.cpp */
//#include <stdio.h>
//#include <stdlib.h>
//#include <math.h>
//#include <string.h>

using namespace std;


/***** NSGA-II parameters *****/

#define SLOWSORTING			1
#define FASTSORTING			2
/* toggles the non-dominated sorting method */
#define SORTING SLOWSORTING
/* toggles whether the slow and fast sorting procedures should attempt to produce identical outputs. 
This is not always possible if several solutions in the population end up being identical. 
Setting this to zero is slightly faster. */
#define ATTEMPT_IDENTICAL	1

/***** SBB parameters *****/

/* maximum number of attributes that can be indexed. */
#define INPUTS 65536 /* Requires 16 bits. */
/* Probability of not randomly selecting a specific exemplar. */
#define PFAIL 1e-31

extern map < string, string > args;

/***********************************************************************************************************/

class esc
{
	
public:
	
	/* Data set name. */
	string _dataSetName;
	/* Determines if the given data set provides labels or not. */
	int _labelsProvided;
	/* Label set name. */
//	string _labelSetName;
	/* Dimensionality of the data set. */
	long _dataSetDim;
	/* Cardinality of the data set. */
	long _dataSetSize;
	
	/* Number of objectives being used by the symbionts and hosts. */
	int _numLevels;
	/* Number of objectives used by the symbionts. */
	int _numObjectives0;
	/* Number of objectives used by the hosts. */
	int _numObjectives1;
	/* Number of objectives used by the symbionts or hosts. */
	int _numObjectives;
	
	/* These variables determine if specific implemented objectives should be used or not. */
	int _doDistortion;
	int _doConnectivity;
	int _doAttrCount;
	int _doClusterCount;
	
	/* Whether or not the subsampling is performed. */
	int _doSubsampling;
	/* How often is the sampling done. */
	int _samplingPeriod;
	
	/* Point population size. */
	int _pointPopSize;
	
	/* Vector of indexes to active subset points. */
	vector < int > _activePoints;
	
	/* Host population size. */
	int _hostPopSize;
	
	/* Min and Max number of symbionts used to form a host. */
	int _minSymbionts;
	int _maxSymbionts;
	
	/* Minimum Probability of Crossover rate. */
	double _pc_initial;
	/* Maximum Probability of Crossover rate. */
	double _pc_final;
	/* Minimum Probability of Mutation rate. */
	double _pm_initial;
	/* Maximum Probability of Mutation rate. */
	double _pm_final;

	/* Probability of Crossover rate, calculated based on pc_max and pc_min. */
	double _pc;
	/* Probability of Mutation rate, calculated based on pm_max and pm_min. */
	double _pm;
	/* Probability of symbiont deletion.*/
	double _psd;
	/* Probability of symbiont addition. */
	double _psa;
	/* Probability of symbiont swap. */
	double _pss;
	/* Probability of multi-level mutation. */
	double _pmlm;
	
	/* Min and Max number of attributes used to make a symbiont. */
	int _minAttributes;
	int _maxAttributes;
	
	/* Probability of attribute deletion. */
	double _pad;
	/* Probability of attribute addition. */
	double _paa;
	/* Probability of attribute swap. */
	double _pas;
	/* Probability of location change. */
	double _plc;
	
	/* Maximum number of generations for symbiont and host evolutionary processes. */
	long _numGenerations0;
	long _numGenerations1;
	long _numGenerations;
	
	/* Do we use attribute probability learning? */
	int _useAttrProbLrng;
	
	/* Learning rate for adjusting atrribute probabilities. */
	double _attrProbLrngRate;
	
	/* Random number seed. */
	int _seed;
	
	/* Statistics printing period. */
	int _statMod;
	
	/* Maximum number of tries to get a point of a specific label. */
	long _maxTries;
	
	/* Maximum number of attempts to generate a random number. */
	int _maxAttempts;
	
	/* Epsilon (a very small value.) */
	double _epsilon;
	
	/* Infinity (a very large value.) */
	double _infinity;
	
	/* 1-D cluster centroids' file name. */
	string _attrWiseClusCentFileName;
	
	/* 1-D cluster labels' file name. */
	string _attrWiseClusLabelsFileName;
	
	/* Noneligible attributes file name. */
	string _noneligibleAttrFileName;
	
	/* Noneligible 1-D cluster centroids' file name. */
	string _noneligibleClusFileName;
	
	/* The set of points defining the environment. The point id should be the 
     index of the point in this vector (and the underlying dataset). */
	vector < double * > _pointSet;
	
	/* The set of labels (if they exist) */
	vector < long > _labelSet;
	
	/* Record index of points in _pointSet grouped by label. The dimension
     of _index should be the number of labels. The dimension of *_index[i]
     should be the number of points of label _labels[i]. This is done to
     speed up sampling of minority classes. */
	vector < vector < long > * > _index;
	
	/* Labels sorted into ascending order. */
	vector < long > _uniqueLabels;
	/* Count the number of points of each label at each index (order matching _labels) */
	vector < long > _labelCount;
	/* Map label->index to find a label in _labels and _labelCount and _index. */
	map < long, long > _labelMap;
	
	/* 1-D cluster centroids. */
	vector < double * > _attrWiseClusCent;
	
	/* Number of 1-D centroids in each attribute.*/
	vector < long > _numClusPerAttr;
	
	/* 1-D cluster labels. */
	vector < long * > _attrWiseClusLabels;
	
	/* Noneligible attributes. */
	vector < long > _noneligibleAttr;
	
	/* Noneligible 1-D cluster centroids. */
	vector < long * > _noneligibleClus;
	
	/* Variable determining which Symbionts in level 1 can be indexed in level 2. */
	int _symbsToUse;

	/* Max number of 1-D cluster centroids in one attribute. */
	int _maxClusPerAttr;
		
	/* Max number of symbionts that can be indexed in second level. */
	int _numSymbsToUse;

	/* Attribute probabilities. */
	vector < double > _attrProbs;
	
	/* 1-D centroids probabilities. */
	//vector < double * > _centProbs;
	
	/* Current generation. */
	long _gen;

	/* Current level. */
	int _lev;

public:
	
	//esc() {};
	esc(map < string, string > &);
	~esc();
	
	void get_params();
	void get_data();
	void readDataSet();
	void normalizeDataSet();
	void readAttrWiseClusCent();
	void readAttrWiseClusLabels();
	void readNoneligibleAttr();
	void readNoneligibleClus();
	void RSS();
	
	/* Get number of labels. */
	inline long numLabels() 
	{ return _uniqueLabels.size(); }

	/* Get label at some index in _uniqueLabels. */
	inline long getLabel(long i) 
	{ return _uniqueLabels[i]; }

	/* Get index in _uniqueLabels and _labelCount and _index of some label. */
	long getLabelIndex(long);

	/* Get the number of points of a given label. */
	inline long getLabelCount(long label) 
	{ return _labelCount[getLabelIndex(label)]; }

	/* Get number of points in a data set. */
	inline long dataSetSize() 
	{ return _pointSet.size(); }

	/* Get the dimensionality of a data set. */
	inline long dataSetDim()
	{ return _dataSetDim; }

	inline void setNumSymbsToUse(int temp)
	{ _numSymbsToUse = temp; }
	
	//void go(map < string, string > &);
	void go();

//	friend class gene;
//	friend class solution;
//	friend class population;
};

#endif
