
#ifndef TYPES_HPP
#define TYPES_HPP

#include "esc.hpp"
#include "misc.hpp"

#include <vector>
#include <iostream>
#include <fstream>

using namespace std;

/***********************************************************************************************************/

class gene
{
public:
	
	// Defines if this gene belongs to a Symbiont (type==0) or a Host(type==1) solution.
	int level;
	
	// Depending on the type, index1 and index2 encode the related indexes to attributes, or symbionts.  
	vector <int> index1;
	vector <int> index2;
	
	// Determines if solution is modified since last generation (by crossover or mutation)
	bool modified;
	
	gene(int);//int
	
	gene() { /*cout << "default gene constructor!" << endl;*/ }
//	~gene() { cout << "default gene destructor!" << endl; }
	
	int print(int);

	void randomize();
	
	inline int size()
	{ return (int) index1.size();}
	
	void pruneDuplicates();
	bool pruneSymbsWithZeroInsts();
	
	void sort();
	
	bool isEligible();

	bool isModified()
	{return modified;}
	
	void setModified(bool mdfd)
	{modified = mdfd;}
	
	// Checks if two genes are the same
	bool operator== (const gene&) const;
	// A substitue for operator==
	//bool equals(const gene&) const;
	
	void crossover( gene& p1,  gene& p2);
	void crossover_level0_1point( gene& p1,  gene& p2); // for ind type 0
	void crossover_level0_exchange( gene& p1,  gene& p2); // for ind type 0
	void crossover_level1_1point( gene& p1,  gene& p2); // for ind type 1

	void mutation();
	void mutation_level0_removeAttr(); // for symbionts of type == 0
	void mutation_level0_addAttr(); // for symbionts of type == 0
	void mutation_level0_replaceAttr(); // for symbionts of type == 0
	void mutation_level0_replaceClus(); // for symbionts of type == 0
	void mutation_level1_removeSymb(); // for hosts of type != 0
	void mutation_level1_addSymb(); // for hosts of type != 0
	void mutation_level1_replaceSymb(); // for hosts of type != 0
	
};

/***********************************************************************************************************/

class solution
{
private:
	
	// Determines if the solution is a Symbiont (type==0), or a Host (type==1). 
	int level;
	
	gene sol;
	
	// Determines if solution is modified since last generation (by crossover or mutation)
	bool modified;
	
	// Objective vector when using MOGA.
	vector <double> objectives;
	
	// overall sum of the shared fitness;
	//static double static_sum;
	
	// Single fitness value when using fitness sharing.
	double fitness;
	
	// Crowding Distance used in NSGA-II algorithm
	double crowdingdistance;
	
	// Front Number used in NSGA-II algorithm
	// Originally in public section.
	unsigned int frontnumber;
	
	// Number of times this solution has been inedxed in higher level individuals.
	int refs;
	
	// Attributes used by the solution
	vector <int> attrs;
	
	// Instances belonging to the solution prior to the final nearest neighbour
	vector <int> insts;
	
	// Distances between this symbiont and each point.
	vector <double> distances;
	
	double attrPerSymb;
	
public:
	
	//solution() { cout << "default solution constructor!" << endl; }
	solution();
	solution(solution *);
	~solution();
	
	int print(int);
	
	void randomize();
	
	vector <int> evaluate(int);
	
	bool isEligible()
	{return sol.isEligible(); }
	
	bool isModified()
	{return sol.isModified(); }
	
	void setModified(bool mdfd)
	{sol.setModified(mdfd); }
	
	void crossover(solution *far, solution *mor);
	
	void mutation();
	
	int getLevel()
	{ return level; }
	
	int getRefs()
	{ return refs; }
	
	gene getGene()
	{ return sol; }
	
	void setGene1(vector <int> tempvec)
	{ sol.index1 = tempvec; }
	
	void setGene2(vector <int> tempvec)
	{ sol.index2 = tempvec; }
	
	int size()
	{ return (int) sol.size(); }
	
	void calcDistances();
	
	vector <double> getDistances()
	{ return distances; }
	
	vector <double> getObjectives()
	{ return objectives; }
	
	double getObj(int o) const
	{ return objectives[o]; }
	
	double getFitness()
	{ return fitness; }
	
	int getFrontNumber()
	{ return frontnumber; }
	
	void setFrontNumber(int f)
	{ frontnumber = f; }
	
	double getCrowdingDistance()
	{ return crowdingdistance; }
	
	void setCrowdingDistance(double cd)
	{ crowdingdistance = cd; }
	
	void resetCrowdDist()
	{ crowdingdistance = 0.0; }
	
	void addtoCrowdDist(double d)
	{ crowdingdistance += d; }
	
	void updateFront(unsigned int f)
	{ frontnumber = max(f,frontnumber); }
	
	void updateFront(solution &s)
	{ frontnumber = max(s.frontnumber+1,frontnumber); }
	
	vector <int> getAttrs()
	{ return attrs; }
	
	void updateAttrs();
	
	vector <int> getInsts()
	{ return insts; }
	
	void increfs()
	{ refs++; }
	
	void decrefs()
	{ refs--; }
	
	/* Checks if one solution dominates another in objective currentobj. */
	bool operator< (const solution&) const;

	/* Checks if one solutions has better frontnumber/crowddistance than another. */
	bool operator<< (const solution&) const;

	/* Checks if one solution dominates another. */
	bool operator>> (const solution&) const;

	/* Checks if two solutions have identical genes. */
	bool operator== (const solution&) const;
	
	/* Checks if one solution dominates another in the first M objectives. */
	bool MdomA(const solution &, unsigned int M);
	bool MdomB(const solution &, unsigned int M);

	bool twoDcmp(const solution&) const;

	bool cmpOnCD(const solution& s2) const
	{ return(crowdingdistance > s2.crowdingdistance); }

	void performance_measures(int do_print, int do_print2);

#if SORTING==SLOWSORTING
	int domcount;
	// list is the obvious choice, but vector is faster.
	vector <solution *> dominates;
#endif

};

/***********************************************************************************************************/

class front
{
private:
	
	// Determines if the front is formed by Symbionts (type==0), or Hosts (type==1). 
	int level;
	
	// Number of the front
	int frontnumber;
	
	// Pointers/Indexes to solutions on a front
	vector <solution *> indexes;

	// Copy of the pointers/indexes to solution on a front
	vector <solution *> backup_indexes;

	// Pointers/Indexes to unique solution on a front (no redundant solutions)
	vector <solution *> unique_indexes;
	
	vector <int> uniqueIndexes;

public:

	front();

	front(int);

	front(solution *, int);

	int print(int);
	
	// Returns i'th solution of the front.
	solution& operator[] (int i)
	{ return(*indexes[i]); }
	
	vector <solution *> getIndexes()
	{ return indexes; }
	
	vector <int> getUniqueIndexes()
	{ return uniqueIndexes; }
	
	void swap(front &sf);

	void calcCrowdingDistances();

	void sortOnCD();

	// Used to be in private section
	void sortOnObjective(int o); 
	
	void push_back(solution *s)
	{
		indexes.push_back(s);
		s->setFrontNumber(frontnumber);
	}

	vector <solution *>::size_type size()
	{ return(indexes.size()); }

//	vector <int> uniqueSolutionsIndexes();

//	vector <int> paretoFrontLabelAssignment();

};

/***********************************************************************************************************/

class population
{
private:
	
	// Determines if the population is formed by Symbionts (type==0), or Hosts (type==1). 
	int level;
	
	unsigned int popsize, doublepopsize;

	vector <solution> pop;

	vector <solution *> popsorted;

	vector <front> sortedfronts;
	
	front paretofront;
	
	vector <int> attrs_on_PF;
	vector <int> insts_on_PF;

	//  void sort_on_fronts();

#if SORTING==FASTSORTING
	// procedures for fast non-dom sorting:
	void twoDsort();

	void sort_on_objective(int o); 

	void ND_helper_A(vector<solution *> &S, unsigned int M,bool dosort);

	void ND_helper_B(vector<solution *> &L, vector<solution *> &H, unsigned int M, bool dosort);
#endif

public:
	
	population();//int
	
	int print(int);
	
	vector <solution> getPop()
	{ return pop; }

	void randomize();

	void evaluate(int);
	
	int size()
	{ return popsize; }
	
	const solution& operator[] (int i)
	{ return(*popsorted[i]); }

	void sortOnFronts(); // private!

	void calcCrowdingDistances();

	void breedGeneration(int);
	
	void finalizeGeneration(int, int);
	
	void finalizeLevel(int);

	void attrsInstsOnPF();

//	front getParetoFront();

	vector <front> getSortedFronts()
	{ return(sortedfronts); }
	
	vector <solution *> getPopsorted()
	{ return(popsorted); }
	
	int bestOnFitness();
	
	void run();

};


#endif
